// Level Set advection and reinitialisation solver for
// C-LSM: Conservative Level Set Method
//
// Created: 2006-09-13
// Dag Lindbo, dag@f.kth.se

#include "LSM_Solver.hpp"

using namespace dolfin;

LSM_Solver::LSM_Solver(Mesh& mesh, Function& vf, real epsi, real k)
  : mesh(mesh), vf(vf), dt(k)
{
  // SETTING INTERNAL PARAMETERS

  // Reinit params
  iter_re_max = 100;

  // Div by zero in grad(phi) normalization
  normal_cutoff = 1e-9;

  // INITIALIZE THE LSM SOLVER
  epsilon = epsi;
  dtau = dt;
  min_cell = mf.minDiam(mesh);
  diff_coeff = 0.0001*min_cell;

  // Functions
  phi = new Function(x, mesh, element_phi);
  phi_0 = new Function(x0, mesh, element_phi);
  phi_norml = new Function(norml_vec, mesh, element_grad);

  // Forms
  a_adv = new LSM_adv_CN::BilinearForm(vf,dt);
  L_adv = new LSM_adv_CN::LinearForm(*phi_0, vf, dt);

  a_re = new LSM_re::BilinearForm(*phi_0, *phi_norml, dtau, epsilon);
  L_re = new LSM_re::LinearForm(*phi_0, *phi_norml, dtau, epsilon);

  // Assembly of constant LHS
  FEM::assemble(*a_adv, A_adv, mesh);
}

// ADV-REINIT TIMESTEP ASSUMING VEL. FIELD TO BE FIXED, dt CONST
Vector& LSM_Solver::timestep(Vector& xn)
{
  x0 = xn;
  tau = 0;
  iter_re = 0;

  FEM::assemble(*L_adv, b_adv, mesh);
  lu_solver.solve(A_adv, x, b_adv);

  x = df.computeSmoothing(mesh, *phi, diff_coeff);
  norml_vec = df.computeUnitNormal(mesh, *phi);

  while(iter_re < iter_re_max)
    {
      x0 = x;
      FEM::assemble(*a_re,*L_re, A_re, b_re, mesh);
      lu_solver.solve(A_re, x, b_re);

      iter_re++;
      tau +=dtau;
      
      err_rel = vff.computeRelError(x, x0);
      dolfin_info("err_rel = %f",err_rel);
      if(err_rel < 0.1*dt)
	break;
    }
  dolfin_info("\t\t\t\t Iter re : %d",iter_re);
  return x;
}

// ADV-REINIT TIMESTEP WITH UPDATED VEL. FIELD
Vector& LSM_Solver::timestep(real k, Vector& xn, Function& vel)
{
  vf = vel;
  x0 = xn;
  dt = k;
  tau = 0;
  dtau = dt;
  iter_re = 0;

  FEM::assemble(*a_adv,*L_adv, A_adv, b_adv, mesh);
  lu_solver.solve(A_adv, x, b_adv);

  x = df.computeSmoothing(mesh, *phi, diff_coeff);
  norml_vec = df.computeUnitNormal(mesh, *phi);

  while(iter_re < iter_re_max)
    {
      x0 = x;
      FEM::assemble(*a_re,*L_re, A_re, b_re, mesh);
      lu_solver.solve(A_re, x, b_re);

      iter_re++;
      tau +=dtau;
      
      err_rel = vff.computeRelError(x, x0);
      dolfin_info("err_rel = %f",err_rel);
      if(err_rel < 0.1*dt)
	break;
    }
  dolfin_info("\t\t\t\t Iter re : %d",iter_re);
  return x;
}

// REINIT LOOP ONLY
Vector& LSM_Solver::timestep_only_reinit(Vector& xn,real tol)
{
  x = xn;
  tau = 0;
  iter_re = 0;

  x= df.computeSmoothing(mesh, *phi, diff_coeff);
  norml_vec = df.computeUnitNormal(mesh, *phi);

  Progress p("Reinitializing");

  while(iter_re < iter_re_max)
    {
      x0 = x;
      FEM::assemble(*a_re,*L_re,A_re,b_re,mesh);
      lu_solver.solve(A_re, x, b_re);

      iter_re++;
      tau +=dtau;
      
      err_rel = vff.computeRelError(x, x0);
      p = iter_re/(double)iter_re_max;

      dolfin_info("err_rel = %f",err_rel);
      if(err_rel < tol*dt)
	break;
    } 
  dolfin_info("\t\t\t\t Iter re : %d",iter_re);
  return x;
}

void LSM_Solver::solve()
{

}
