// 
// C-LSM: Conservative Level Set Method
//
// Test bench forvarious LSM components
// 
//
// Created: 2006-09-30
// Dag Lindbo, dag@f.kth.se

#include <dolfin.h>
#include <dolfin/timing.h>

#include "Vec_Func_fcns.hpp"
#include "forms/LSM_adv_CN.h"
#include <fstream>

using namespace dolfin;

// INITIAL CONDITIONS -------------------------------------------%%%
// INITIAL CONDITION: CIRCULAR BUBBLE
class IC_bubble : public Function
{
public:

  real IC_epsi, IC_radius, IC_cen_x, IC_cen_y;

  IC_bubble(real epsi, 
	    real rad, 
	    real cen_x, 
	    real cen_y)
  {
    IC_epsi = epsi;
    IC_radius = rad;
    IC_cen_x = cen_x;
    IC_cen_y = cen_y;
  }

  real eval(const Point& p, unsigned int i)
  {
    real x = p.x;
    real y = p.y;
 
    return 1/(1+std::exp(1/IC_epsi*(std::sqrt((x-IC_cen_x)*(x-IC_cen_x)+
	     (y-IC_cen_y)*(y-IC_cen_y) )-IC_radius)));
  }
};

// VELOCITY FIELD: "ROTATE"
class Vel_Field_Rotate : public Function
{
public:
  
  Vel_Field_Rotate() : Function((unsigned int)2) {}

  real eval(const Point& p, unsigned int i)
  {
    if(i == 0)
      return p.y;
    else if(i == 1)
      return -p.x;
    else return 0;
  }
};

// BEGIN EXECUTION ----------------------------------------------%%%
int main(void)
{
  tic();

  // Params
  int N = 32;

  // Initial condition parameters
  real IC_epsi = 0.01; 
  real IC_radius = 0.25; 
  real IC_cen_x = 0.5; 
  real IC_cen_y = 0.5; 

  // Operators
  Vec_Func_fcns vff;

  // Initial conditions
  IC_bubble IC_fcn(IC_epsi,IC_radius,IC_cen_x,IC_cen_y);

  // Mesh
  UnitSquare mesh(N, N);

  // FEM components
   LSM_adv_CN::BilinearForm::TrialElement element_phi;

  // Misc
  File file_phi("output/LSM_phi.pvd");

  // Const velocity field
  Vel_Field_Rotate vf;

  // Level set func
  Vector x;
  Function phi(x, mesh, element_phi);

  // Map IC to DF on mesh
  x = vff.restrictUFtoDF(mesh,IC_fcn);

  file_phi << phi;

  // Testing error computations

  // Change in this function to make program work
  real err1 = vff.L2_Error_Exact(mesh, IC_fcn, phi); // This returns fine

  // Testing area computation
  real area1 = vff.integrateLevelSet(mesh, phi); // SIGSEGV here
  real area2 = vff.integrateLevelSet2(mesh, phi);

  return 0;
}
